//
// <PW-Id>
//

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright 2018-2020 by Pierre Wolfers Meylan France                       //
//                                                                           //
// This library is free software. Distribution and use rights are outlined   //
// in the file "COPYING" which should have been included with this file.     //
// If this file is missing or damaged, see the license at:                   //
//                                                                           //
//    <To be define when ready>                                              //
//                                                                           //
//   This license described in this file overrides all other licenses that   //
//   might be specified in other files for this library.                     //
//                                                                           //
//   This library is free software; you can redistribute it  and/or modify   //
//   it under the terms of the GNU Lesser General Public License as publi-   //
//   shed by  the  Free Software  Foundation;  either  version 2.1  of the   //
//   License, or (at your option) any later version.                         //
//                                                                           //
//   This library  is  distributed in  the  hope that  it will  be useful,   //
//   but   WITHOUT  ANY  WARRANTY;  without even  the  implied warranty of   //
//   MERCHANTABILITY  or  FITNESS  FOR A  PARTICULAR PURPOSE.  See the GNU   //
//   Library General Public License for more details.                        //
//                                                                           //
//   You should have  received  a copy of  the  GNU Lesser General  Public   //
//   License  along with this library  (see COPYING.LIB); if not, write to   //
//   the Free Software Foundation :                                          //
//                        Inc., 675 Mass Ave, Cambridge, MA 02139, USA.      //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



//
// P.Wolfers DiaViewer Software
//

//
// Module to manage a slide collection
// DCM = DiaViewer Collection Management.
//

#include <FL/Fl.H>
#include <FL/fl_ask.H>
#include <FL/fl_utf8.h>         // To manipulate utf8 strings.

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <string.h>

/*
#if defined(WIN32) && !defined(__CYGWIN__)
# include <dir.h>
# define MKDIR( p ) mkdir( p )
#else
# define MKDIR( p ) mkdir( p, 0 )
#endif
*/

// #include <bits/stdc++.h> 
// using namespace std; 

#include "SetUpMan.h"
#include "DiaViewer_DIA.h"
#include "DiaViewer_DCM.h"
#include "DiaViewer_PRG.h"
#include "DiaViewer_CTX.h"




//
// Each diapositive/slide is identified by a unique IIde_t integer.
// This identifier include three part. The volume identifier number,
// the directory identifier number (in the volume) and the slide identifier
// unique in the volume. The three field in the IIde_t identifier can be 
// defined by a definition as :
// #define LARGE_UNI_IDENT
// to have a 64 bits identifier type (the default is 32 bits), and by
// #define UNIQUE_IDE_FIELDS {<sld>,<dir>}
// For example, for 32 bits identifiers:
// #define  UNIQUE_IDE_FIELDS {16,10}
// while by used to get a slide field of 16 bits and a directories field of 10 bits,
// that keep 6 bits (32 - 16 - 10 = 6) for the volume. With this, you can have a
// maximum of 2**16-1 = 65535 slides in each directory, 2**10-1 = 1023 directories
// in each volume and 2**(32-16-10)-1 = 64 volumes. The default 32 bits IIde Identifier
// is defined at the compilation time but the UNIQUE_IDE_FIELDS is keep in the user's
// local data context and can be changed when the DiaViewer software is restarted and
// it should be possible to change these field sizes by using the setup menu in some
// future versions.
// The default fields size values are :
//    #define UNIQUE_IDE_FIELDS {14,10} in 32 bits that allows
//           16,383 slides in each directory,
//             1023 directories in each volume and
//              255 opened volumes.
//
//    #define UNIQUE_IDE_FIELDS {32,16} in 64 bits that allows
//    4,294,967,297 slides in each directory,
//           65,535 directories in each volume and
//           65,535 opened volumes.
//
// The default DiaViewer software generation is with Unique Identifiers of 32 bits.
//
//



//
// Our Task'Global Variables Definitions (public).
//
const char  DirSep  =   DIRSEP; // Separator to use between files and directories for our OS.


//
// Global variables.
//


int         SlideNumber   =  0, // Current Slide number in its film.
            FilmNumber    =  0, // Current film number.
            ChapterNumber =  0, // Current Chapter Number.
            VolumeNumber  =  0; // Current Volume Number.

VolEntry  * IniSelVolum = NULL; // Selected volume pointer.
SliEntry  * IniSelSlide = NULL; // Initial selected Slide pointer.


//
// Our module specific Variables Definitions (private).
//

static FilePath   BltPath( 0 ); // To keep the current path During the construction of volume tree.

/*
static int    max_nbdir =    0, // Maximul number of Sub-Directories in a Directory.
              max_nbsli =    0; // Maximum number of slides in a Directory.
*/

static int    curr_deep =    0, // Current directory deep from the base directory.
              curr_tsli =    0, // Current total numbers of slides and subdirectories.
              curr_tdir =    0,
              maxi_deep =    0, // Maximum directory deep.
              deep_error   = 0; // Slide Volume overlay error flag.

/// static char * CFilePath = NULL; // Config File Path (when found).

/// static FILE *             cfgf; // Configuration file.



#define PRT( f, v ) printf( f, v ); fflush( stdout )

//
// Entry type defined to directory entries.
//

typedef void * Entry;

typedef struct { // Define the table of table (for sorting).
    int    use,  // Used size.
           all;  // Allocated size.
    Entry *tab;  // The table of name.
} EntryTable;

static int DirFlg = 0; // for sorting directories/slides entries.



//
// Some routines.
//

static EntryTable * NewEntry( Entry en, EntryTable *tb = NULL )
{  // Allocate a enough large Sort table.
    Entry *tm;
    int  size;

    if (!tb) {
        tb = new EntryTable;
        tb->use =    0;
        tb->all =    0;
        tb->tab = NULL;
        size    =    0;
    }
    size = tb->use + 1;
    if (size >= tb->all) {
        size = (size + TABLE_ALLOC_STEP)&~(TABLE_ALLOC_STEP - 1);
        tm = new Entry[size];
        if (tb->use)
            for(int ii=0;ii<tb->use;ii++) tm[ii] = tb->tab[ii];
        if (tb->tab) delete[] tb->tab;
        tb->tab =   tm;
        tb->all = size;
    }
    tb->tab[tb->use++] = en;
    return tb;
} // static EntryTable * NewEntry( Entry en, EntryTable *tb = NULL ).



static void FreeEntryTable( EntryTable * etb )
{
    etb->use = etb->all = 0;
    if (etb->tab) delete[] etb->tab;
    delete etb;
} // static void FreeEntryTable( EntryTable * etb ).



static int CmpItemName( Entry *tb, int jp, Entry id )
{ // Compare the jp'th entry name in the tb entry table with the string sp. 
    const char * sj = (DirFlg)? ((DirEntry*)tb[jp])->FName() : ((SliEntry*)tb[jp])->FName(),
               * sp = (DirFlg)? ((DirEntry*)id)->FName() : ((SliEntry*)id)->FName();

//#if defined(WIN32) || defined(__EMX__)
    return fl_utf_strcasecmp( sj, sp );
//#else
//    return fl_utf_strcasecmp( sj, id, nnnn );
//#endif
} // static int CmpItemName( Entry *tb, int jp, Entry id ).



static void ItemSimplSort( Entry *tb, int min, int max )
{
    int     jj;
    Entry   id;
    
    for(int ii=min+1; ii<=max; ii++) {
        id = tb[ii];
        jj = ii - 1;
        while (jj>=min&&CmpItemName( tb, jj, id )>=0) {
            tb[jj+1] = tb[jj];
            jj--;
        }
        tb[jj+1] = id;
    }
} // static void ItemSimplSort( Entry *tb, int min, int max ).



static void ItemQuickSort( Entry *tb, int min, int max )
// Perform a Quick sort.
{
    int     ic, ip;
    Entry   pi, tm;
    
    if (min>=max) return;
    
    ip = (min + max)/2;         // Get the middle Sort table index ...
    pi =        tb[ip];         // ... as the pivot.

    tb[ip] = tb[min];           // Put the pivot item in first position.
    tb[min] = pi;               //

    ic = min;                   //
    for(int ii = min; ii <= max; ii++)
    { // Loop on all entries.
        if (CmpItemName( tb, ii, pi )<0) {
            if (ii != ++ic) {
                // When the entry are differents, we exchange them.
                tm = tb[ii]; tb[ii] = tb[ic]; tb[ic] = tm;
            }
        }
    }
    // Put the middle entry to its right place.
    tm = tb[ic]; tb[ic] = tb[min]; tb[min] = tm;

    if (ic-min<QSORT_LIMIT) ItemSimplSort( tb, min, ic-1 );
                       else ItemQuickSort( tb, min, ic-1 );
    if (max-ic<QSORT_LIMIT) ItemSimplSort( tb, ic+1, max );
                       else ItemQuickSort( tb, ic+1, max );
} // static void ItemQuickSort( Entry *tb, int min, int max ).



static void EntrySort( Entry *tb, int size )
{
    if (size<QSORT_LIMIT) ItemSimplSort( tb, 0, size-1 );
                     else ItemQuickSort( tb, 0, size-1 );
} // static void EntrySort( Entry *tb, int size ).



static SliEntry * SliInList( char * name )
{
    SliEntry   * nwsli;
    int     ii, jj, ll;
    char        ext[8];

    const char extab[] = "jpg ,jpeg,JPG ,JPEG,png ,PNG ,bmp ,BMP ,0000";
    const int  n_extab = 8;
    
    // We must check for supported image file.
    ii = strlen( name ); ll = 0;
    while (name[--ii] != '.' && ll < 5) ll++;
    if (ll>4||ll<3) return NULL;                // Unknown file extension -> skip file.
    
    ext[0] = name[++ii];                        // Get the founded file extention.
    ext[1] = name[++ii];
    ext[2] = name[++ii];
    ext[3] = (ll==4)?name[++ii]:' ';
    ext[4] = 0;
    jj=0;
    for (ii=0;ii<n_extab;ii++) {                // Loop to find when it a supported file type.
        if (!memcmp( ext, extab+jj, 4 )) break;
        jj += 5;
    }
    if (ii>=n_extab) return NULL;               // Unknown file extension -> skip file.
    
    // Now we can include this image in the current directory image list.

    nwsli = new SliEntry( name, 0 );
    
/// printf( " Enter the image \"%s\"\n", dia->diafnam );
    
    return nwsli;
} // static SliEntry * SliInList( char * name ).



static int OvrlTreeTest()
{
    int pthl, ierr;

    pthl = BltPath.Length();
    BltPath.Set( CNTX_FNAME, 0 );
    ierr = CntxOpen( BltPath.Path(), 0 );
    if (ierr > 0) {
        // The Context file exist.
        
        CntxClose();
    }
    BltPath.Rem( pthl );

    return ierr;
} // static int OvrlTreeTest().



static DirEntry * ScanFoundDir( const char * name ) {
    // pdir is the parent directory descriptor,
    // path is the complete path where is founded the new directory entry,
    // name is the name of the new directory entry.
    //

    DIR          * dir; // Directory file descriptor used by opendir(),
                        // readir() and closedir().
    struct dirent *dsc; // Directory entry descriptor - from readir().
    struct stat  entbf; // File or Sub-directory descriptor - from stat().
    DirEntry   * nwdir; // Pointer to the current directory descriptor,
    SliEntry   * nwsli; // Pointer to the current diapositive descriptor.
    EntryTable * lidir, // Sub directory pointers list.
               * lisli; // Slides/Diapositives pointers list.
    int      ncdir = 0, // Cumullated number of sub-directories found.
             ncsli = 0, // Cumullated number of diapositive found.
             ndir  = 0, // Local number of sub-directories found.
             nsli  = 0; // Local number of diapositives found.

    int        basplen; // Keep the base Path length.

    BltPath.Set( NULL, 0 );             // Open Directory without final DirSep.

#ifdef _HAVE_FL_UTF8_HDR_
    if (!(dir=opendir( BltPath.Path() ))) {
#else
    if (!(dir=opendir( BltPath.Path() ))) {
#endif
        fprintf( stderr, "\n  * DiaViewer Cannot open the directory :\n %s \n * %s => Scan Abort\n",
                 BltPath.Path(), /* PathString, */ strerror( errno ) );

        BltPath.Set( NULL, 1 );         // Restore the Directory separator.
        return NULL;
    }

    BltPath.Set( NULL, 1 );   // Re-append the final "/" and keep length to restore original path...
    basplen = BltPath.Length();

//printf( " Our Path = (%d)\"%s\"\n", BltPath.Length(), BltPath.Path() );

    lidir       = NULL; // Initialize the sub-directories and ...
    lisli       = NULL; // ... diapositives list.


// Scan this directory.
    while (dsc=readdir(dir)) {          // Read a directory entry.
        if (dsc->d_name[0]!='.') {      // Ignore all entry as ., .. and special dir/files.
            // Complete the path to get access to the entry (File or Directory).
            BltPath.Set( dsc->d_name, 0 );

//printf( " New path <%s>\n", BltPath.Path() );

            // Check for directory entry category (skip when no file or directory).
#if defined( _WIN32 ) && !defined( _CYGWIN_ )
            if (stat( BltPath.Path(), &entbf ))
#else
            if (lstat( BltPath.Path(), &entbf ))
#endif
                // One entry access error.
                printf( " (l)stat error %s\n on entry \"%s\"\n",
                        strerror( errno ), BltPath.Path() );

            else {      // Entry Access is OK.
                switch (entbf.st_mode & S_IFMT) {
                    case S_IFDIR: { // For directory.
                        BltPath.Set( NULL, 1 );
                        curr_deep++;
                        if (nwdir = ScanFoundDir( dsc->d_name )) {
                            // Append this directory in the directory list.
                            lidir = NewEntry( (void*)nwdir, lidir );
                            // Update the diapositive and directory counts.
                            ncdir += nwdir->NDir() + 1;
                            ncsli += nwdir->CnSli();
                            ndir++;
                        }
                        curr_deep--;
                    break;
                    }

                    case S_IFREG: { // For regular file.
                        if (nwsli = SliInList( dsc->d_name )) {
                            // Put the new diapositive in the directory diapositive list.
                            lisli = NewEntry( (void*)nwsli, lisli );
                            nsli++;
                            ncsli++;
                        }
                    break;
                    }
                    default: ; // Ignored entry.
                }
            }
            BltPath.Rem( basplen );     // Restore Scanned directory path with its "/".
//          printf( " New path <%s>\n", BltPath.Path() );
        }
    }

//printf( " fin dir scan\n" ); fflush( stdout );
    closedir( dir );

    // Do not keep any directory without diapositives inside (can be in subdirectories).
    
    if (ncsli) {
        DirFlg = 1; // To a directory sort.
        if (ndir>1) EntrySort( lidir->tab, ndir ); // Directories sorting.
        DirFlg = 0; // To a slides sort.
        if (nsli>1) EntrySort( lisli->tab, nsli ); // Slides sorting.

        nwdir = new DirEntry( name, 0 );

        if (lidir) {
            nwdir->SbDir( lidir->use, (DirEntry**)(lidir->tab) );
            FreeEntryTable( lidir );
        }
        if (lisli) {
            nwdir->Slide( lisli->use, (SliEntry**)(lisli->tab) );
            FreeEntryTable( lisli );
        }
        
        nwdir->LDeep( curr_deep );
        nwdir->CnSli( ncsli );

        if (!curr_deep) {
            curr_tdir = ncdir + 1;   
            curr_tsli = ncsli;
        }
        else
        {
            if (curr_deep > maxi_deep) maxi_deep = curr_deep;
        }

        // Try to read a set all directory and slide context informations (old version file).
        BltPath.Set( CNTX_FNAME, 0 );
//printf( " Try to get context\n" ); fflush( stdout );
        deep_error = nwdir->ReadCntxFile( BltPath.Path() );

        BltPath.Rem( basplen );
//printf( " Read context file done for \"%s\".\n", BltPath.Path() ); fflush( stdout );
    }
    else nwdir = NULL;
//printf( " ZZZ\n" ); fflush( stdout );
//PRT( " v Fin %d\n", curr_deep );
    return nwdir;
} // static DirEntry * ScanFoundDir( const char * name ).



static VolEntry * CreateVolume( const char * name,
                                DirEntry   * root,
                                int        newdir )

// Function to create a new Volume of slides (diapositives).
// The name argument is the volume root directory.
// If <newdir> is true (!=0), a new volume directory is created
// with the path <name>. else the directory access is checked
// for read access.
//
{
    int            flg = 1;
    int            idv = 0;
    VolEntry  * vol = NULL;
	
    if (newdir) { // We create a new slides volume directory.
        flg = fl_mkdir( name, 0 );
    } else { // We use an existing slides volume directory.
        flg = fl_access( name, R_OK );
    }
		
    if (!flg) {
        // The creator of VolEntry put the Volume reference in the table OpenVolTable.
        vol = new VolEntry( name, root, idv );          // Allocate the Volume Entry and ...

        // Now, complete the volume initialization.

        vol->VRoot( root );
        vol->TDeep( maxi_deep );
        vol->TnDir( curr_tdir );
        vol->TnSli( curr_tsli );

    } else {
        vol = NULL;
    }
    maxi_deep = 0;      // Reset the deep maximum for a next volume creation.
    curr_tsli = 0;
    curr_tdir = 0;
    return vol;
} // CreateVolume( const char * name, DirEntry * root, int newdir ).



static void SetVolumeIdents( VolEntry *vol )
// set the Unique integer identifier for a complete volume.
{
    if (vol) {
        vol->SetVolIdents();    // Setup the volume identifier.
        if (IIdeOverflow)  {
            if (IIdeOverflow&1) fl_alert( "\n  * DiaViewer : Too many slides.\n" );
            if (IIdeOverflow&2) fl_alert( "\n  * DiaViewer : Too many slides'directories.\n" );
            if (IIdeOverflow&4) fl_alert( "\n  * DiaViewer : Too many slides'volumes.\n" );
            // Actions correctrices !!!
        }
    }
} // static void SetVolumeIdents( VolEntry *vol ).



static int CreateVolumeTree( const char * dirfnam, int indx )
// Scan and create the diapositives (Slides) tree list.
// The total number of found diapositives is returned (<0 on error or 0 for empty slide volume).
// The procedure return this OpenVolTable index of the volume or -1 on fatal error.
//
{
    DirEntry     * root;
    VolEntry     *  vol;

    BltPath.Clear();            // Force new path to be built.
    BltPath.Set( dirfnam, 1 );  // Set PathString initial value for a Directory.
    maxi_deep              = 0; // Initialize the directory maximum deep.

    root = ScanFoundDir( dirfnam );     // Scan the directory.
    // Here the PathString is restored at dirfnam initial value.

//  printf( " DCM fdir Root? %d\n", root? 1 : 0 );

    if (root) {
        vol = CreateVolume( dirfnam, root, 0 );

        vol->SetVolContext();   // Set the Volume specific data.

        SetVolumeIdents( vol ); // Set all vol/dir/slide integer identifiers.

        SlidePrg * prg = new SlidePrg( vol );

        prg->Display();

        vol->RelPrg( (void*)prg );
    } else vol = NULL;

    BltPath.Clear();            // Clear the Path.

    if (deep_error) {
        // Positive value : The volume include an other one..
        // Negative value : The volume is include in an other one.
        fl_alert(
            " *** DiaViewer Overlay Error on the volume :\n Path = \"%s\"\n Read only use recommanded",
            vol->VRoot()->FName()
        );
    }
    return (root)? OpenVolTable.TabUse() : -1;
} // static int CreateVolumeTree( const char * dirfnam, int indx ).



static void CleanWindowsTable( int idx )
// To suppress the window of volume reference # idx.
//
{
    int ip = 0;

    if (WinTable.size) {
        while ((ip < WinTable.size)&&(idx != WinTable.tab.i[ip])) ip += 4;
        if (ip < WinTable.size) { // The related windows id found.
            ip += 4;
            while (ip < WinTable.size) {
                WinTable.tab.i[ip-4] = WinTable.tab.i[ip];
                ip++;
            }
            WinTable.size -= 4;
        }
        if (!WinTable.size) { delete[] WinTable.tab.i; WinTable.tab.i = 0; }
    }
} // static void CleanWindowsTable( int idx ).



void DisplayAllImageTree();


void OpenVolumeQuList()
// Scan the volume list VolQueue (old) or VolToOpen table (new - in priority)
// and open each.
{
    int               idx, idv;
    VolEntry     * svol = NULL;
    const char *          path;

    if (VolToOpen.size) {
        int      nvol = VolToOpen.size;

        for(int i = 0; i < nvol; i++) {
            path = VolToOpen.tab.s[i];
            idv = CreateVolumeTree( path, -1 );
            if (idv < 0) {
                // Send an Alert message and remove the Volume.
                fl_alert( "Open Slide Volume Error %d : \n path = %s", -idv, path );
                CleanWindowsTable( i );
            }
            else VolToOpen.tab.s[i] = NULL;
        }
        delete[] VolToOpen.tab.s;
        if (SetUpQuNoEmpty( VolQueue )) SetUpQuErase( VolQueue, 0 );
    } else {
        idx = SetUpQuLocate( VolQueue,  -1 );
        while (idx) {
            path = SetUpQuGetS( VolQueue );
            idv = CreateVolumeTree( path, idx );
            if (idv < 0) {
                // Send an Alert message and remove the Volume.
                fl_alert( "Open Slide Volume Error %d : \n path = %s", -idv, path );
                CleanWindowsTable( idx );
            }
            idx = SetUpQuLocate( VolQueue, 0 );
        }
        SetUpQuErase( VolQueue, 0 );      // Erase the queue.
    }

    IniSelSlide = NULL;
    
//  DisplayAllImageTree();
} // void OpenVolumeQuList().



int OpenSlideVolume( const char *fpath )
// To Open the appropriate Current slide volume.
// Path is the path to directory where find slides.
//
{
    // Now create the Slide Volume tree structures.
    return CreateVolumeTree( fpath, 0 );

} // int OpenSlideVolume( const char *fpath, int dflg ).





//
// **********************************************************************************
//


static void PrtSpace( int n ) {
    for (int ii=0; ii<n; ii++) putchar( ' ' );
} // static void PrtSpace( int n ).



void DisplayDirectory( DirEntry * dir, int nsp )
{
    DirEntry             * sdi;
    SliEntry             * sli;
    int    ndir, nsli, nbr = 0;

    if (dir) {
        ndir = dir->NDir();
        nsli = dir->NSli();
    
        PrtSpace( nsp ); printf( "Directory %8xH/ \"%s\" (lvl=%d) :\n",
                                 dir->Ident(), dir->FName(), dir->LDeep() );
        if (dir->SText()) { PrtSpace( nsp ); printf( "* title \"%s\"\n", dir->SText() ); }
        PrtSpace( nsp ); printf( "* with %d subdirectorie(s) and %d slide(s) or a total number of %d slide(s) \n",
                                 ndir, nsli, dir->CnSli() );
        // Edit the list of diapositives.
        nsp += 2;

        for(int i=1;i<=ndir;i++) {
            sdi = dir->SbDir( i );
            if (sdi) DisplayDirectory( sdi, nsp );
        }
        for(int i=1;i<=nsli;i++) {
            sli = dir->Slide( i );
            if (sli) {
                PrtSpace( nsp );
                printf( "%3d:%8xH/ %s\n", nbr++, sli->Ident(), sli->FName() );
                printf( " * fl %d, or1 %d, or2 %d, sn %d, fn %d, pn %d, cn %d\n",
                        sli->FlgVal(), sli->OriOpe(), sli->CurOpe(), sli->Numb(),
                        sli->Film(), sli->Para(), sli->Chap() );
            }
        }
        PrtSpace( nsp-2 ); printf( "****\n" );
    }
    else
        printf( "/n *** error: The directory pointer is NULL ***\n" );
} // void DisplayDirectory( DirEntry * dir, int nsp ).



void DisplayImageTree( VolEntry* vol )
{
    printf( "\n List of Diapositives Tree \"%s\":\n\n", BltPath.Path() );

    if (vol) {
        const char * vnam = vol->VName();

        if (vnam) printf( " Volume Name \"%s\"\n", vnam );

        DisplayDirectory( vol->VRoot(), 1 );
		
        printf( "\n There are a total of %d slides in %d directories\n\n", vol->TnSli(), vol->TnDir() );

        printf( "***\n\n" );
    } else printf( "\n Slide Volume was not existing.\n" );
} // void DisplayImageTree( VolEntry* vol ).



void DisplayAllImageTree()
{
    for(int i = 0; i < OpenVolTable.TabUse(); i++) {
        DisplayImageTree( (VolEntry*)(OpenVolTable.GetObj( i )) );
    }
} // void DisplayAllImageTree().




//
// end of <PW-Id>.
//
