//
// <PW-Id>
//

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright 2018-2020 by Pierre Wolfers Meylan France                       //
//                                                                           //
// This library is free software. Distribution and use rights are outlined   //
// in the file "COPYING" which should have been included with this file.     //
// If this file is missing or damaged, see the license at:                   //
//                                                                           //
//    <To be define when ready>                                              //
//                                                                           //
//   This license described in this file overrides all other licenses that   //
//   might be specified in other files for this library.                     //
//                                                                           //
//   This library is free software; you can redistribute it  and/or modify   //
//   it under the terms of the GNU Lesser General Public License as publi-   //
//   shed by  the  Free Software  Foundation;  either  version 2.1  of the   //
//   License, or (at your option) any later version.                         //
//                                                                           //
//   This library  is  distributed in  the  hope that  it will  be useful,   //
//   but   WITHOUT  ANY  WARRANTY;  without even  the  implied warranty of   //
//   MERCHANTABILITY  or  FITNESS  FOR A  PARTICULAR PURPOSE.  See the GNU   //
//   Library General Public License for more details.                        //
//                                                                           //
//   You should have  received  a copy of  the  GNU Lesser General  Public   //
//   License  along with this library  (see COPYING.LIB); if not, write to   //
//   the Free Software Foundation :                                          //
//                        Inc., 675 Mass Ave, Cambridge, MA 02139, USA.      //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



//
// P.Wolfers DiaViewer Software
//

//
// Module of service routines/objects.
//


#include <string.h>
#include <FL/fl_utf8.h>         // To manipulate utf8 strings.

#include "DiaViewer_SRV.h"

// Must be powers of 2.
#define VSTR_ALLOC_STEP 16
#define PATH_ALLOC_STEP 64
#define TABL_ALLOC_STEP 16


//
// Functions for Temporary string allocations/free class VarString.
//

char * VarString::Size( int size )
{   char *nst_;
    use_ = size + 1;            // Add one for the final null character.
    if (all_ < use_) {          // We must increase the string allocation.
        all_ = (use_ + VSTR_ALLOC_STEP)&~(VSTR_ALLOC_STEP-1);
        nst_ = new char[all_];  // Allocate the new string.
        // Copy the old string in the new one when required.
        if (cop_&&str_&&str_[0]) strcpy( nst_, str_ );
                            else nst_[0] = 0;
        if (str_) delete[] str_;// Free old string.
        str_ = nst_;
    }
    return str_;
} // void VarString::Allocate( int size ).



//
// **********************************************************************************
//


//
// Class FilePath to manage a dynamic file path.
//


FilePath        * CPath = NULL; // The CPath FilePath must be visible out of any class.


FilePath::FilePath( int stks )
{
    if (stks > 0) stks++;
    sl_  = stks;             // To execute stks request without stack overflow.
    path_ = NULL;
    length_ = 0;
    allsiz_ = 0;
    if (sl_ > 0) stack_ = new int[sl_];
            else stack_ = NULL;
    sp_ = -1;
} // FilePath::FilePath( int stks ).



FilePath::~FilePath()
{
    if (path_) delete[] path_;
    if (stack_) delete [] stack_;
} // FilePath::~FilePath().



void FilePath::StackSize( int stks )
{
    if (stks >= 0) {
        stks++;
        if (stks >= sl_) {
            delete[] stack_;
            stack_ = new int[stks];
        }
        sl_ = stks; sp_ = -1;
    } else {
        if (stack_) delete[] stack_;
        sl_ = 0;
    }
} // void FilePath::StackSize( int siz ).



int FilePath::Set( const char * name, int dfl )
//
// Set a string in *PathString to be the path of the new file of directory.
// SetFilePath return the old PathString length to be reset to old value by
// the routine RemFilePath( int ).
// When the argument name is NULL the PathString is preserved.
// The argument dfl is used to set or remove a terminated Directory separator
// in all case, (name argument provided of NULL).
//
{
    int len_, oln_, usz_;
    
    oln_ = length_;
    if (name) {
        len_ = strlen( name );
        usz_ = len_ + length_ + 2;
        if (allsiz_ < usz_) {
            // Allocated size round up to a multiple of STRING_ALLOC_STEP.
            allsiz_ = (usz_ + PATH_ALLOC_STEP)&~(PATH_ALLOC_STEP-1); 
            char * tmp_ = new char[allsiz_];
            if (path_) {
                strcpy( tmp_, path_ );
                delete[] path_;
            }
            path_ = tmp_;
        }
        usz_ = length_ + len_;
        strcpy( path_ + oln_, name );
        if (stack_) {
            if (sp_ < sl_) stack_[++sp_] = oln_;
                      else oln_ = -1;
        }
    } else usz_ = length_;
    
    if (path_[usz_-1]!=DIRSEP) {
        if (dfl) {
            path_[usz_++] = DIRSEP;
            path_[usz_] = 0;
        }
    } else {
        if (!dfl) {
            path_[--usz_] = 0;
        }
    }
    length_ = usz_;
   return oln_;
} // int FilePath::Set( const char * name, int dfl ).



int FilePath::Rem( int vl )
{
    int ie = 0;

    if (stack_) {
        if (sp_ >= 0) length_ = stack_[sp_--];
                 else ie = -1;
    } else length_ = vl;
    path_[length_] = 0;
    return ie;
} // int FilePath::Rem( int vl ).



void FilePath::Clear()
{
    if (path_) path_[0] = 0;
    length_ =  0;
    sp_     = -1;
} // void FilePath::Clear().



//
// Object table class routines to manage a tabel of undefined objects.
//

void ObjTable::AddObj( ObjPtr p )
{
    ObjPtr *tm;
    int   size;

    if (!p) return;
    size = use_ + 1;
    if (size >= all_) {
        size = (size + TABL_ALLOC_STEP)&~(TABL_ALLOC_STEP - 1);
        tm = new ObjPtr[size];
        if (use_)
            for(int ii = 0; ii < use_; ii++) tm[ii] = tab_[ii];
        if (tab_) delete[] tab_;
        tab_ =   tm;
        all_= size;
    }
    tab_[use_++] = p;
    return;
} // void ObjTable::AddObj( ObjPtr p ).



ObjPtr * ObjTable::GetObjs( int shf )
{
    if ((shf >= 0)&&(shf < use_))
        return (ObjPtr*)((char*)tab_ + shf );
    else
        return NULL;
} // ObjPtr * ObjTable::GetObjs( int shf )



void ObjTable::Tamp( int idx )
{
    int  i,  j;

    if ((idx < 0)||(idx >= use_)) return;

    if (tab_) {
        i = j = idx;
        while (i < use_) {
            if (tab_[i]) {
                if (j != i) tab_[j] = tab_[i];
                i++; j++;
            } else i++;
        }
        use_ = j;
    }
} // void ObjTable::Tamp().



void ObjTable::RemObj( int idx )
{
    if ((idx >= 0)&&(idx < use_)) {
        tab_[idx] = NULL;
        this->Tamp( idx );
    }
} // void ObjTable::RemObj( int idx ).



void ObjTable::RemObj( ObjPtr p )
{
    int idx = 0;

    while ((idx < use_)&&(tab_[idx] != p)) idx++;
    if (idx < use_) {
        tab_[idx] = NULL;
        this->Tamp( idx );
    }
} // void ObjTable::RemObj( ObjPtr p ).



//
// end of <PW-Id>.
//
